import java.awt.Color;
import java.awt.geom.Point2D;
import java.text.DecimalFormat;
import java.util.Iterator;

import ch.aplu.util.*;
import ch.beattl.gm3d.*;
import ch.beattl.projektor.*;

//Projiziert 3D-Objekte auf den Bildschirm
public class Projektor3D_v05 extends GPanel
{
	static final double Radius = 0.025; // Radius des Eckpunktes
	static final double Epsylon = 0.0001; // Ab wann ist x=0?
	static final double zoomfactor = Math.sqrt(2.); // Zoomfaktor
	static final double dalpha = Math.PI / 20; // Winkel, um den gedreht wird
	// static final String fileName = "Tetraeder.obj";
	static final String fileName = "Wuerfel.obj";
	// static final String fileName = "Oktaeder.obj";
	// static final String fileName = "Ikosaeder.obj";
	// static final String fileName = "Dodekaeder.obj";

	Polyeder polyeder; // Variable fuer das darszustellende Polyeder
	Point Auge; // Position des Auges
	boolean nummernFlag = false; // true=Nummern der Ecken einblenden,
	// false=Ausblenden
	boolean kantenFlag = true; // true=Ecken einblenden, false=Ausblenden

	/**
	 * Projiziert den Punkt P auf die Bildebene Pi (xy-Ebene) und
	 * gibt den Bildpunkt zurueck.
	 * @param P Punkt P
	 * @return Bildpunkt
	 */
	Point2D.Double projectOnPi(Point P)
	{
		// Parameterwert bis zur Projektionsebene Pi
		double t = P.z - Auge.z;
		if (Math.abs(t) > Epsylon)
			t = -Auge.z / t;
		// R ist der Durchstosspunkt des Strahls (Auge,P) durch Pi
		double x = Auge.x + t * (P.x - Auge.x);
		double y = Auge.y + t * (P.y - Auge.y);
		Point2D.Double R = new Point2D.Double(x, y);

		// Rueckgabewert
		return R;
	}

	/**
	 * Zeichnet die Eckpunkte des Wuerfels in einer GPanel-Graphik.
	 */
	void draw()
	{
		// Alle Ecken zeichnen
		for (int i = 0; i < polyeder.getNumOfVertices(); i++)
		{
			// Durchstosspunkt <Auge,Ecke> mit Pi berechnen
			Point2D.Double R = projectOnPi(polyeder.getVertex(i));

			// Punkt R zeichnen und beschriften
			color(Color.red);
			move(R);
			fillCircle(Radius);
			if (nummernFlag)
			{
				color(Color.black);
				move(R.x + .05, R.y);
				text("" + (i + 1));
			}
		}
		// Alle Kanten zeichnen
		if (kantenFlag)
		{
			Iterator<Face> it = polyeder.getFaces();
			while (it.hasNext())
			{
				Face flaeche = it.next();
				// Polygon zusammensetzen
				int k = flaeche.getNumOfVertices();
				Point2D.Double[] polygon = new Point2D.Double[k];
				for (int j = 0; j < k; j++)
				{
					// Nummer der Ecke finden
					int n = flaeche.getVertexN(j);
					// Ecke auf Bildebene projizieren
					polygon[j] = projectOnPi(polyeder.getVertex(n));
				}
				color(Color.blue);
				polygon(polygon);
			}
		}
		// Infos auf den Bildschirm
		showMenu(-2.25, 2.25);
		showParameters(1.25, 2.25);
	}

	/**
	 * Skaliert das Polyeder mit dem Faktor factor.
	 * @param factor Skalierungsfaktor
	 */
	void scale(double factor)
	{
		Matrix m = new Matrix(
				factor, 0, 0,
				0, factor, 0,
				0, 0, factor);
		for (int i = 0; i < polyeder.getNumOfVertices(); i++)
		{
			Point P = polyeder.getVertex(i);
			m.transform(P);
		}
	}
	
	/**
	 * Dreht das Polyeder um den Winkel alpha um die Drehachse axis
	 * @param phi Drehwinkel
	 * @param axis Drechachse
	 */
	public void rotate(double phi, char axis)
	{
		Matrix m;
		switch (axis)
		{
			case 'x':
				m = new Matrix(
						1, 0, 0,
						0, Math.cos(phi), -Math.sin(phi),
						0, Math.sin(phi), Math.cos(phi));
				break;
			case 'y':
				m = new Matrix(
						Math.cos(phi), 0, Math.sin(phi),
						0, 1, 0,
						-Math.sin(phi), 0, Math.cos(phi));
				break;
			case 'z':
				m = new Matrix(
						Math.cos(phi), -Math.sin(phi), 0,
						Math.sin(phi), Math.cos(phi), 0,
						0, 0, 1);
				break;
			default:
				m = new Matrix();
		}
		for (int i = 0; i < polyeder.getNumOfVertices(); i++)
		{
			Point P = polyeder.getVertex(i);
			m.transform(P);
		}
	}

	/**
	 * Zeigt ein Menue fuer die Tastatureingabe an der Position (x / y).
	 * @param x x-Koordinate der Position
	 * @param y y-Koordinate der Position
	 */
	void showMenu(double x, double y)
	{
		double delta = 0.15;
		color(Color.black);
		move(x, y);
		text("r - Alles zuruecksetzen");
		move(x, y - 1 * delta);
		text("+/- - Ein- / Auszoomen");
		move(x, y - 2 * delta);
		text("n - Eckennr. " + (nummernFlag ? "= true" : "= false"));
		move(x, y - 3 * delta);
		text("k - Kanten " + (kantenFlag ? "= true" : "= false"));
		move(x, y - 4 * delta);
		text("x,s - Drehung um x-Achse");
		move(x, y - 5 * delta);
		text("y,a - Drehung um y-Achse");
		move(x, y - 6 * delta);
		text("c,d - Drehung um z-Achse");
		move(x, y - 7 * delta);
		text("ESC - Beenden");
	}

	/**
	 * Zeigt einige Parameter zur Darstellung an der Position (x / y).
	 * @param x x-Koordinate der Position
	 * @param y y-Koordinate der Position
	 */
	void showParameters(double x, double y)
	{
		double delta = 0.15;
		color(Color.black);
		move(x, y);
		text("Objekt: " + polyeder.getName());
		move(x, y - 1 * delta);
		DecimalFormat df = new DecimalFormat("##0.#");
		text("Auge = (" + df.format(Auge.x) + "/" + df.format(Auge.y) + "/"
				+ df.format(Auge.z) + ")");
	}

	Projektor3D_v05()
	{
		// Graphikfenster definieren
		super("Graphik", -2.5, 2.5, -2.5, 2.5);

		// Augposition festlegen
		Auge = new Point(0, 0, 5);

		// Objektdefinition
		FileReader myReader = new FileReader(fileName);
		polyeder = new Polyeder(myReader);
		polyeder.setName(fileName);
	}

	public static void main(String[] args)
	{
		Projektor3D_v05 proj = new Projektor3D_v05();

		int key;
		do
		{
			// Polyeder zeichnen
			proj.draw();

			// Tasten Abfragen
			key = proj.getKeyWait();

			switch (key)
			{
				case '+':
					proj.clear();
					proj.scale(zoomfactor);
					break;
				case '-':
					proj.clear();
					proj.scale(1./zoomfactor);
					break;
				case 'n':
					proj.clear();
					proj.nummernFlag = !proj.nummernFlag;
					break;
				case 'k':
					proj.clear();
					proj.kantenFlag = !proj.kantenFlag;
					break;
				case 'r':
					proj.clear();
					FileReader myReader = new FileReader(fileName);
					proj.polyeder.loadData(myReader);
					break;
				case 'x':
					proj.clear();
					proj.rotate(dalpha, 'x');
					break;
				case 's':
					proj.clear();
					proj.rotate(-dalpha, 'x');
					break;
				case 'y':
					proj.clear();
					proj.rotate(dalpha, 'y');
					break;
				case 'a':
					proj.clear();
					proj.rotate(-dalpha, 'y');
					break;
				case 'c':
					proj.clear();
					proj.rotate(dalpha, 'z');
					break;
				case 'd':
					proj.clear();
					proj.rotate(-dalpha, 'z');
					break;
			}
		} while (key != 27);
		System.exit(0);
	}
}
